#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'

import os
import shutil
import struct

import numpy as np
import pandas as pd

from tqsdk.channel import TqChan
from tqsdk.diff import _get_obj
from tqsdk.rangeset import _rangeset_difference, _rangeset_intersection
from tqsdk.tafunc import get_dividend_df, get_dividend_factor
from tqsdk.utils import _generate_uuid

CACHE_DIR = os.path.join(os.path.expanduser('~'), ".tqsdk/data_series")


class DataSeries:
    """
    获取数据模块，支持用户获取任意时间段内的 Kline / Tick 数据，返回格式为 pandas.DataFrame

    支持用户设置开始时间和结束时间，且返回的 df 不会随着行情更新

    首个版本功能点：
    1. 缓存数据，目录 ~/.tqsdk/data_series，文件名规则为 symbol.dur.start_id.end_id                                                                          
    2. 每次下载数据先检查缓存数据是否包含需要下载数据，计算出还需要下载的数据段
    3. 如果有需要下载的数据段
       下载的数据先写在 symbol.dur.temp 文件，下载完成之后文件名修改为 symbol.dur.start_id.end_id
    4. 下载完数据之后，合并连续的数据文件，更新缓存文件
    5. 如果有复权参数，计算复权数据
    6. 返回完整的 pandas.DataFrame 给用户


    **功能限制说明**
    * 该接口返回的 df 不会随着行情更新
    * 暂不支持多合约 Kline 缓存
    * 不支持用户回测/复盘使用。get_data_series() 是直接连接行情服务器，是会下载到未来数据的。
    * 不支持多进程/线程/协程。每个合约+周期只能在同一个线程/进程里下载，因为需要写/读/修改文件，多个线程/进程会造成冲突。

    """

    def __init__(self, api, symbol_list, dur_nano, start_dt_nano, end_dt_nano, adj_type=None) -> None:
        """
        创建历史数据下载器实例

        Args:
            api (TqApi): TqApi实例，该下载器将使用指定的api下载数据

            symbol_list: 需要下载数据的合约代码，当指定多个合约代码时将其他合约按第一个合约的交易时间对齐

            dur_nano (int): 数据周期，纳秒数

            start_dt_nano (int): 起始时间, 纳秒数

            end_dt_nano (int): 结束时间, 纳秒数

            adj_type (str/None): 复权计算方式，默认值为 None。"F" 为前复权；"B" 为后复权；None 表示不复权。只对股票、基金合约有效。
        """
        self._api = api
        self._symbol_list = symbol_list if isinstance(symbol_list, list) else [symbol_list]
        self._dur_nano = dur_nano
        self._start_dt_nano = start_dt_nano
        self._end_dt_nano = end_dt_nano + 1  # 实际上用户需要下载的时间段使用 Range 表示为 [self.start_dt_nano, self.end_dt_nano)
        self._adj_type = adj_type
        self._dividend_cache = {}  # 缓存合约对应的复权系数矩阵，每个合约只计算一次
        self.df = pd.DataFrame()
        self.is_ready = False
        DataSeries._ensure_cache_dir()  # 确认缓存文件夹存在
        self._api.create_task(self._run())

    async def _run(self):
        symbol = self._symbol_list[0]  # todo: 目前只处理一个合约的情况
        # 检查缓存文件，计算需要下载的数据段
        rangeset_id = DataSeries._get_rangeset_id(symbol, self._dur_nano)
        rangeset_dt = DataSeries._get_rangeset_dt(symbol, self._dur_nano, rangeset_id)
        DataSeries._assert_rangeset_asce_sorted(rangeset_id)
        DataSeries._assert_rangeset_asce_sorted(rangeset_dt)
        # 计算差集 todo: 已下载数据的 rangeset_id[0][0] ~ rangeset_id[-1][1] 覆盖到的部分，可以使用 id 范围下载，未覆盖到的使用 dt 范围下载
        # rangeset_dt 中的最后一个 range 的最后一根 k 线，需要重新下载，因为如果用户填写的 end_dt_nano 是未来时间，有可能这跟 k 线数据是过时的
        if len(rangeset_dt) > 0:
            rangeset_dt[-1] = (rangeset_dt[-1][0], rangeset_dt[-1][1] - self._dur_nano)
            if rangeset_dt[-1][0] == rangeset_dt[-1][1]:
                # 如果最后一个 range 的只包含最后一根 k 线，则删去最后一个 Range，最后一根 k 线需要重新下载
                rangeset_dt.pop(-1)
        diff_rangeset = _rangeset_difference([(self._start_dt_nano, self._end_dt_nano)], rangeset_dt)

        # 下载数据并全部完成
        if len(diff_rangeset) > 0:
            await self._download_data_series(diff_rangeset)
            self._merge_rangeset()  # 归并文件
            rangeset_id = DataSeries._get_rangeset_id(symbol, self._dur_nano)
            rangeset_dt = DataSeries._get_rangeset_dt(symbol, self._dur_nano, rangeset_id)
            DataSeries._assert_rangeset_asce_sorted(rangeset_id)
            DataSeries._assert_rangeset_asce_sorted(rangeset_dt)

        assert len(rangeset_id) == len(rangeset_dt) > 0

        # 查找用户请求时间段与已有数据时间段的交集
        target_rangeset_dt = _rangeset_intersection([(self._start_dt_nano, self._end_dt_nano)], rangeset_dt)
        assert len(target_rangeset_dt) <= 1  # 用户请求应该落在一个时间段内，或者用户请求的时间段内没有任何数据
        if len(target_rangeset_dt) == 0:  # 用户请求的时间段内没有任何数据
            self.is_ready = True
            return

        # 此时用户请求时间范围，转化为 target_rangeset_dt[0]
        # 找到 self._start_dt_nano, self._end_dt_nano 对应的 range_id
        (start_dt, end_dt) = target_rangeset_dt[0]
        range_id = None
        for index in range(len(rangeset_dt)):
            range_dt = rangeset_dt[index]
            if range_dt[0] <= start_dt and end_dt <= range_dt[1]:
                range_id = rangeset_id[index]
                break
        assert range_id

        # 返回 dataframe, 目标文件为
        filename = os.path.join(CACHE_DIR, f"{symbol}.{self._dur_nano}.{range_id[0]}.{range_id[1]}")
        data_cols = DataSeries._get_data_cols(symbol=self._symbol_list[0], dur_nano=self._dur_nano)
        dtype = np.dtype([('id', 'i8'), ('datetime', 'i8')] + [(col, 'f8') for col in data_cols])
        fp = np.memmap(filename, dtype=dtype, mode='r', shape=range_id[1] - range_id[0])

        # target_rangeset_dt[0] 对应的 id 范围是 [start_id, end_id)
        start_id = fp[fp['datetime'] <= start_dt][-1]["id"]
        end_id = fp[fp['datetime'] < end_dt][-1]["id"]
        rows = end_id - start_id + 1  # 读取数据行数
        array = fp[start_id - range_id[0]: start_id - range_id[0] + rows]

        # 按列赋值，df.append 方法会返回一个新的 df，初始化构造的 df 可能（如果是异步代码下）已经返回给用户代码中；df.update 方法会导致更新的列的 type 都是 object
        self.df["id"] = array["id"]
        self.df["datetime"] = array["datetime"]
        for c in data_cols:
            self.df[c] = array[c]
        self.df["symbol"] = symbol
        self.df["duration"] = self._dur_nano

        # 复权, 如果存在 STOCK / FUND 并且 adj_type is not None, 这里需要提前准备下载时间段内的复权因子
        quote = self._api.get_quote(symbol)
        if self._adj_type and quote.ins_class in ["STOCK", "FUND"]:
            factor_df = await self._update_dividend_factor(symbol)  # 复权需要根据日线计算除权因子，todo: 如果本地下载过日线，不需要再从服务器获取日线数据
            if self._adj_type == "F":
                # 倒序循环 factor_df, 对于小于当前 factor_df[datetime] 的行 乘以 factor_df[factor]
                for i in range(factor_df.shape[0] - 1, -1, -1):
                    dt = factor_df.iloc[i].datetime
                    factor = factor_df.iloc[i].factor
                    adj_cols = DataSeries._get_adj_cols(symbol, self._dur_nano)
                    lt = self.df["datetime"].lt(dt)
                    self.df.loc[lt, adj_cols] = self.df.loc[lt, adj_cols] * factor
            if self._adj_type == "B":
                # 正序循环 factor_df, 对于大于等于当前 factor_df[datetime] 的行 乘以 1 / factor_df[factor]
                for i in range(factor_df.shape[0]):
                    dt = factor_df.iloc[i].datetime
                    factor = factor_df.iloc[i].factor
                    adj_cols = DataSeries._get_adj_cols(symbol, self._dur_nano)
                    ge = self.df["datetime"].ge(dt)
                    self.df.loc[ge, adj_cols] = self.df.loc[ge, adj_cols] / factor
        # 结束状态
        self.is_ready = True

    async def _download_data_series(self, rangeset):
        symbol = self._symbol_list[0]
        for start_dt, end_dt in rangeset:
            try:
                start_id, end_id = None, None
                temp_filename = os.path.join(CACHE_DIR, f"{symbol}.{self._dur_nano}.temp")
                temp_file = open(temp_filename, "wb")
                data_chan = TqChan(self._api)
                task = self._api.create_task(self._download_data(start_dt, end_dt, data_chan))
                async for item in data_chan:
                    temp_file.write(struct.pack("@qq" + "d" * (len(item) - 2), *item))
                    if start_id is None:
                        start_id = item[0]
                    end_id = item[0]
            except Exception as e:
                temp_file.close()  # 这里如果遇到 exception，不应该再重命名文件，相当于下载失败
            else:
                temp_file.close()
                if start_id is not None and end_id is not None:
                    target_filename = os.path.join(CACHE_DIR, f"{symbol}.{self._dur_nano}.{start_id}.{end_id + 1}")
                    shutil.move(temp_filename, target_filename)
            finally:
                task.cancel()
                await task

    async def _download_data(self, start_dt, end_dt, data_chan):
        # 下载的数据应该是 [start_dt, end_dt）
        symbol = self._symbol_list[0]
        chart_info = {
            "aid": "set_chart",
            "chart_id": _generate_uuid("PYSDK_data_series"),
            "ins_list": symbol,
            "duration": int(self._dur_nano),
            "view_width": 2000,
            "focus_datetime": int(start_dt),
            "focus_position": 0,
        }
        # 还没有发送过任何请求, 先请求定位左端点
        self._api._send_chan.send_nowait(chart_info)
        chart = _get_obj(self._api._data, ["charts", chart_info["chart_id"]])
        current_id = None  # 当前数据指针
        path = ["klines", symbol, str(self._dur_nano)] if self._dur_nano != 0 else ["ticks", symbol]
        serial = _get_obj(self._api._data, path)
        cols = DataSeries._get_data_cols(symbol, self._dur_nano)
        try:
            async with self._api.register_update_notify() as update_chan:
                async for _ in update_chan:
                    if not (chart_info.items() <= _get_obj(chart, ["state"]).items()):
                        continue  # 当前请求还没收齐回应, 不应继续处理
                    left_id = chart.get("left_id", -1)
                    right_id = chart.get("right_id", -1)
                    if (left_id == -1 and right_id == -1) or self._api._data.get("mdhis_more_data", True):
                        continue  # 定位信息还没收到, 或数据序列还没收到
                    # 检查合约的数据是否收到
                    if serial.get("last_id", -1) == -1:
                        continue
                    if current_id is None:
                        current_id = max(left_id, 0)
                    while current_id <= right_id:
                        item = serial["data"].get(str(current_id), {})
                        if item.get("datetime", 0) == 0 or item["datetime"] >= end_dt:
                            return  # 当前 id 已超出 last_id 或k线数据的时间已经超过用户限定的右端
                        row = [current_id, item["datetime"]] + [DataSeries._get_float_value(item, c) for c in cols]
                        await data_chan.send(row)
                        current_id += 1
                        self._current_dt_nano = item["datetime"]
                    # 当前 id 已超出订阅范围, 需重新订阅后续数据
                    chart_info.pop("focus_datetime", None)
                    chart_info.pop("focus_position", None)
                    chart_info["left_kline_id"] = current_id
                    await self._api._send_chan.send(chart_info)
        finally:
            # 释放chart资源
            await data_chan.close()
            await self._api._send_chan.send({
                "aid": "set_chart",
                "chart_id": chart_info["chart_id"],
                "ins_list": "",
                "duration": self._dur_nano,
                "view_width": 2000,
            })

    async def _update_dividend_factor(self, symbol):
        quote = self._api.get_quote(symbol)
        df = get_dividend_df(quote.stock_dividend_ratio, quote.cash_dividend_ratio)
        between = df["datetime"].between(self._start_dt_nano, self._end_dt_nano)  # 只需要开始时间～结束时间之间的复权因子
        df["pre_close"] = float('nan')
        for i in df[between].index:
            chart_info = {
                "aid": "set_chart",
                "chart_id": _generate_uuid("PYSDK_data_factor"),
                "ins_list": symbol,
                "duration": 86400 * 1000000000,
                "view_width": 2,
                "focus_datetime": int(df.iloc[i].datetime),
                "focus_position": 1
            }
            await self._api._send_chan.send(chart_info)
            chart = _get_obj(self._api._data, ["charts", chart_info["chart_id"]])
            serial = _get_obj(self._api._data, ["klines", symbol, str(86400000000000)])
            try:
                async with self._api.register_update_notify() as update_chan:
                    async for _ in update_chan:
                        if not (chart_info.items() <= _get_obj(chart, ["state"]).items()):
                            continue  # 当前请求还没收齐回应, 不应继续处理
                        left_id = chart.get("left_id", -1)
                        right_id = chart.get("right_id", -1)
                        if (left_id == -1 and right_id == -1) or self._api._data.get("mdhis_more_data",
                                                                                     True) or serial.get("last_id",
                                                                                                         -1) == -1:
                            continue  # 定位信息还没收到, 或数据序列还没收到, 合约的数据是否收到
                        last_item = serial["data"].get(str(left_id), {})
                        # 复权时间点的昨收盘
                        df.loc[i, 'pre_close'] = last_item['close'] if last_item.get('close') else float('nan')
                        break
            finally:
                await self._api._send_chan.send({
                    "aid": "set_chart",
                    "chart_id": chart_info["chart_id"],
                    "ins_list": "",
                    "duration": 86400000000000,
                    "view_width": 2
                })
        df["factor"] = (df["pre_close"] - df["cash_dividend"]) / df["pre_close"] / (1 + df["stock_dividend"])
        df["factor"].fillna(1, inplace=True)
        return df

    def _merge_rangeset(self):
        symbol = self._symbol_list[0]
        rangeset = DataSeries._get_rangeset_id(symbol, self._dur_nano)
        if len(rangeset) <= 1:
            return
        # 可以 merge 的 rangset 分组, rangset_group 类型是 list_of_list, [[(start, end, rows), ], [], ....]
        # 内层列表的每个元素值为 range 的 start, end，用来找到正确的文件名，rows 表示 merge 文件时，应该从当前一段 range 里 merge 多少行
        # 主要用于有一种情况：已有 rangeset 的最后一根 kline，有可能被重新下载，导致某个待 merge 的 rangeset 形式为 [0, 50), [49, 50)
        # 此时，[10, 50), [49, 50) 两个 range 应该 merge 到最终文件里的行数 rows 是 49，1; 而不是 50，1。
        # 实际上除了这种情况之外，一个 range 应该 merge 到最终文件里的行数都应该是 range.end - range.start
        # 所以在生产 rangset_group 时，需要把要 merge 的行数也记下来，方便后面 merge。
        rangset_group = [[rangeset[0] + (rangeset[0][1] - rangeset[0][0], )]]
        for i in range(1, len(rangeset)):
            last_r = rangeset[i - 1]
            r = rangeset[i]
            assert (r[0] < r[1]) & (last_r[0] < last_r[1])
            if i == len(rangeset) - 1:
                # 最后一组 rangeset，last_r[1] 有可能大于 range[0]，因为每次下载过的数据都会认为最后一个数据还是需要重新下载
                assert last_r[1] - 1 <= r[0]
                if last_r[1] == r[0]:
                    rangset_group[-1].append(r + (r[1] - r[0], ))
                elif last_r[1] - 1 == r[0]:
                    # 后一个 range 的 start_id 和与 前一个 range 的 end_id - 1 相等，前一个 range 的最后一个数据被重复下载
                    # 此时的前一个 range 的行数需要 -1
                    rangset_group[-1][-1] = last_r + (last_r[1] - 1 - last_r[0], )
                    rangset_group[-1].append(r + (r[1] - r[0], ))
                else:
                    rangset_group.append([r + (r[1] - r[0], )])
            else:
                assert last_r[1] <= r[0]  # rangeset 应该是严格递增的
                if last_r[1] == r[0]:
                    rangset_group[-1].append(r + (r[1] - r[0], ))
                else:
                    rangset_group.append([r + (r[1] - r[0], )])

        data_cols = DataSeries._get_data_cols(symbol, self._dur_nano)
        dtype = np.dtype([('id', 'i8'), ('datetime', 'i8')] + [(col, 'f8') for col in data_cols])
        for rangeset in rangset_group:
            if len(rangeset) == 1:
                continue
            first_r_0, first_r_1, first_r_rows = rangeset[0]
            # 将第一个文件作为临时文件
            temp_filename = os.path.join(CACHE_DIR, f"{symbol}.{self._dur_nano}.{first_r_0}.{first_r_1}")
            all_rows = first_r_rows  # 临时文件中已经写入的总行数
            last_r_1 = None
            for s, e, rows_number in rangeset[1:]:
                filename = os.path.join(CACHE_DIR, f"{symbol}.{self._dur_nano}.{s}.{e}")
                fp = np.memmap(filename, dtype=dtype, mode='r+', shape=rows_number)
                temp_fp = np.memmap(temp_filename, dtype=dtype, mode='r+', offset=dtype.itemsize*all_rows, shape=rows_number)
                temp_fp[0:rows_number] = fp[0:rows_number]
                temp_fp._mmap.close()
                fp._mmap.close()  # 使用了内部变量 _mmap，fp 还未正式提供 close 方法，参见：https://github.com/numpy/numpy/pull/17941/files
                os.remove(filename)  # 写完后删除源文件
                all_rows += rows_number
                last_r_1 = e
            os.rename(temp_filename, os.path.join(CACHE_DIR, f"{symbol}.{self._dur_nano}.{first_r_0}.{last_r_1}"))

    @staticmethod
    def _assert_rangeset_asce_sorted(rangeset):
        # assert rangeset 是严格升序的，每个 range 都是升序且 前一个range.end < 后一个range.start
        assert all([(rangeset[i][0] < rangeset[i][1]) & (True if i == 0 else rangeset[i-1][1] < rangeset[i][0])
             for i in range(len(rangeset))])

    @staticmethod
    def _get_rangeset_id(symbol, dur_nano):
        # 返回合约 symbol 、周期 dur_nano 的下缓存过的全部 rangeset_id
        rangeset_id = []
        for filename in os.listdir(CACHE_DIR):
            key = f"{symbol}.{dur_nano}."
            if os.path.isfile(os.path.join(CACHE_DIR, filename)) and filename.startswith(
                    key) and "temp" not in filename:
                start_id, end_id = [int(i) for i in filename.split(key)[1].split('.')[-2:]]
                rangeset_id.append((start_id, end_id))
        rangeset_id.sort()  # 对于取到的 rangeset_id 排序， 保证其为升序
        return rangeset_id

    @staticmethod
    def _get_rangeset_dt(symbol, dur_nano, rangeset_id):
        # 返回合约 symbol 、周期 dur_nano 的下 rangeset_id 对应的 rangeset_dt
        rangeset_dt = []
        cols = DataSeries._get_data_cols(symbol, dur_nano)
        dtype = np.dtype([('id', 'i8'), ('datetime', 'i8')] + [(col, 'f8') for col in cols])
        for start_id, end_id in rangeset_id:
            filename = os.path.join(CACHE_DIR, f"{symbol}.{dur_nano}.{start_id}.{end_id}")
            fp = np.memmap(filename, dtype=dtype, mode='r', shape=end_id - start_id)
            first_dt, last_dt = fp[0]["datetime"], fp[-1]["datetime"]
            # 如果是 kline ，range 的 start_dt = first_dt, end_dt = last_dt + dur_nano
            # 如果是 tick， range 的 start_dt = first_dt, end_dt = last_dt + 100
            rangeset_dt.append((first_dt, last_dt + (dur_nano if dur_nano > 0 else 100)))
        return rangeset_dt

    @staticmethod
    def _ensure_cache_dir():
        if not os.path.exists(CACHE_DIR):
            os.makedirs(CACHE_DIR, exist_ok=True)

    @staticmethod
    def _get_data_cols(symbol, dur_nano):
        # 获取数据列
        if dur_nano != 0:
            return ["open", "high", "low", "close", "volume", "open_oi", "close_oi"]
        else:
            cols = ["last_price", "highest", "lowest", "volume", "amount", "open_interest"]
            price_length = 5 if symbol.split('.')[0] in {"SHFE", "SSE", "SZSE"} else 1
            for i in range(1, price_length + 1):
                cols.extend(f"{x}{i}" for x in ["bid_price", "bid_volume", "ask_price", "ask_volume"])
            return cols

    @staticmethod
    def _get_adj_cols(symbol, dur_nano):
        # 复权需要重新计算的列名
        if dur_nano != 0:
            return ["open", "high", "low", "close"]
        else:
            cols = ["last_price", "highest", "lowest"]
            price_length = 5 if symbol.split('.')[0] in {"SHFE", "SSE", "SZSE"} else 1
            cols.extend(f"{x}{i}" for x in ["bid_price", "ask_price"] for i in range(1, price_length + 1))
            return cols

    @staticmethod
    def _get_float_value(obj, key) -> float:
        if key not in obj or isinstance(obj[key], str):
            return float("nan")
        return float(obj[key])


#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'ringo'

from tqsdk import TqApi, TqAuth

api = TqApi(auth=TqAuth("信易账户", "账户密码"))

quote = api.get_quote("SHFE.au2012")

# 预计输出的为以au2012现在最新价来比对的认购的平值期权，当没有符合的平值期权时返回为空,如果有返回则格式为 ["SHFE.au2012C30000"]
ls = api.query_atm_options("SHFE.au2012", quote.last_price, 0, "CALL")

# 预计输出的为au2012，以开盘价来比对的认购的实值3档，实值2档，实值1档期权，如果没有符合要求的期权则对应栏返回为None，如果有则返回格式例如 [None,None,"SHFE.au2012C30000"]
ls = api.query_atm_options("SHFE.au2012", quote.open, [3, 2, 1], "CALL")

# 预计输出的为au2012，以开盘价来比对的认购的实值1档，平值期权，虚值1档，如果没有符合要求的期权则对应栏返回为None，如果有则返回格式例如
ls = api.query_atm_options("SHFE.au2012", quote.open, [1, 0, -1], "CALL")

# 预计输出的为au2012，以现在最新价来比对的认购的虚值1档期权
ls = api.query_atm_options("SHFE.au2012", quote.last_price, -1, "CALL")

# 预计输出沪深300股指期权,2020年12月的虚值1档期权
ls = api.query_atm_options("SSE.000300", quote.last_price, -1, "CALL", exercise_year=2020, exercise_month=12)

# 预计输出 上交所 沪深300股指ETF期权,2020年12月的虚值1档期权
ls = api.query_atm_options("SSE.510300", quote.last_price, -1, "CALL", exercise_year=2020, exercise_month=12)

api.close()



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = "Ringo"

from tqsdk import TqApi, TqAuth

api = TqApi(auth=TqAuth("信易账户", "账户密码"))

# 订阅螺纹钢主连
quote = api.get_quote("KQ.m@SHFE.rb")
# 打印现在螺纹钢主连的标的合约
print(quote.underlying_symbol)

api.close()


#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'

from typing import Dict, Optional

import numpy as np
from pandas import DataFrame, Series

from tqsdk.objs import Account, Trade, SecurityAccount, SecurityTrade
from tqsdk.tafunc import get_sharp, get_sortino, get_calmar, _cum_counts

TRADING_DAYS_OF_YEAR = 250
TRADING_DAYS_OF_MONTH = 21

class TqReport(object):
    """
    天勤报告类，辅助 web_gui 显示回测统计信息和统计图表

    1. 目前只针对 TqSim 账户回测有意义
    2. 每份报告针对一组对应的账户截面记录和成交记录

    """

    def __init__(self, report_id: str, trade_log: Optional[Dict] = None, quotes: Optional[Dict] = None, account_type: str = "FUTURE"):
        """
        本模块为给 TqSim 提供交易成交统计
        Args:
            report_id (str): 报告Id

            trade_log (dict): TqSim 交易结束之后生产的每日账户截面和交易记录
                {
                    '2020-09-01': {
                        "trades": [],
                        "account": {},
                        "positions": {},
                    '2020-09-02': {....},
                }

            quotes (dict): 合约信息

        Example::

            TODO: 补充示例

        """
        self.report_id = report_id
        self.trade_log = trade_log
        self.quotes = quotes
        self.account_type = account_type
        self.date_keys = sorted(trade_log.keys())
        self.account_df, self.trade_df = self._get_df()
        # default metrics
        self.default_metrics = self._get_default_metrics() if self.account_type == "FUTURE" else self._get_stock_metrics()

    def _get_df(self):
        type_account = Account if self.account_type == "FUTURE" else SecurityAccount
        type_trade = Trade if self.account_type == "FUTURE" else SecurityTrade
        account_data = [{'date': dt} for dt in self.date_keys]
        for item in account_data:
            item.update(self.trade_log[item['date']]['account'])
        account_df = DataFrame(data=account_data, columns=['date'] + list(type_account(None).keys()))
        trade_array = []
        for date in self.date_keys:
            trade_array.extend(self.trade_log[date]['trades'])
        trade_df = DataFrame(data=trade_array, columns=list(type_trade(None).keys()))
        if type_trade == Trade:
            trade_df["offset1"] = trade_df["offset"].replace("CLOSETODAY", "CLOSE")
        return account_df, trade_df

    def _get_default_metrics(self):
        if self.account_df.shape[0] > 0:
            result = self._get_account_stat_metrics()
            result.update(self._get_trades_stat_metrics())
            return result
        else:
            return {
                "winning_rate": float('nan'),  # 胜率
                "profit_loss_ratio": float('nan'),  # 盈亏额比例
                "ror": float('nan'),  # 收益率
                "annual_yield": float('nan'),  # 年化收益率
                "max_drawdown": float('nan'),  # 最大回撤
                "sharpe_ratio": float('nan'),  # 年化夏普率
                "sortino_ratio": float('nan'),  # 年化索提诺比率
                "commission": 0,  # 总手续费
                "tqsdk_punchline": ""
            }

    def _get_stock_metrics(self):
        if self.account_df.shape[0] > 0:
            init_asset = self.account_df.iloc[0]['asset_his']
            asset = self.account_df.iloc[-1]['asset']
            self.account_df['profit'] = self.account_df['asset'] - self.account_df['asset'].shift(fill_value=init_asset)  # 每日收益
            self.account_df['is_profit'] = np.where(self.account_df['profit'] > 0, 1, 0)  # 是否收益
            self.account_df['is_loss'] = np.where(self.account_df['profit'] < 0, 1, 0)  # 是否亏损
            self.account_df['daily_yield'] = self.account_df['asset'] / self.account_df['asset'].shift(fill_value=init_asset) - 1  # 每日收益率
            self.account_df['max_asset'] = self.account_df['asset'].cummax()  # 当前单日最大权益
            self.account_df['drawdown'] = (self.account_df['max_asset'] - self.account_df['asset']) / self.account_df['max_asset']  # 回撤
            _ror = asset / init_asset
            return {
                "start_date": self.account_df.iloc[0]["date"],
                "end_date": self.account_df.iloc[-1]["date"],
                "init_asset": init_asset,
                "asset": init_asset,
                "start_asset": init_asset,
                "end_asset": asset,
                "ror": _ror - 1,  # 收益率
                "annual_yield": _ror ** (TRADING_DAYS_OF_YEAR / self.account_df.shape[0]) - 1,  # 年化收益率
                "trading_days": self.account_df.shape[0],  # 总交易天数
                "cum_profit_days": self.account_df['is_profit'].sum(),  # 累计盈利天数
                "cum_loss_days": self.account_df['is_loss'].sum(),  # 累计亏损天数
                "max_drawdown": self.account_df['drawdown'].max(),  # 最大回撤
                "fee": self.account_df['buy_fee_today'].sum() + self.account_df['sell_fee_today'].sum(),  # 总手续费
                "buy_times": self.trade_df.loc[self.trade_df["direction"] == "BUY"].shape[0],  # 买次数
                "sell_times": self.trade_df.loc[self.trade_df["direction"] == "SELL"].shape[0],  # 卖次数
                "max_cont_profit_days": _cum_counts(self.account_df['is_profit']).max(),  # 最大连续盈利天数
                "max_cont_loss_days": _cum_counts(self.account_df['is_loss']).max(),  # 最大连续亏损天数
                "sharpe_ratio": get_sharp(self.account_df['daily_yield']),  # 年化夏普率
                "calmar_ratio": get_calmar(self.account_df['daily_yield'], self.account_df['drawdown'].max()),  # 年化卡玛比率
                "sortino_ratio": get_sortino(self.account_df['daily_yield']),  # 年化索提诺比率
                "tqsdk_punchline": self._get_tqsdk_punchlines(_ror - 1)
            }
        else:
            return {
                "profit_loss_ratio": float('nan'),  # 盈亏额比例
                "ror": float('nan'),  # 收益率
                "annual_yield": float('nan'),  # 年化收益率
                "max_drawdown": float('nan'),  # 最大回撤
                "sharpe_ratio": float('nan'),  # 年化夏普率
                "sortino_ratio": float('nan'),  # 年化索提诺比率
                "fee": 0,  # 总手续费
                "tqsdk_punchline": ""
            }

    def _get_account_stat_metrics(self):
        init_balance = self.account_df.iloc[0]['pre_balance']
        balance = self.account_df.iloc[-1]['balance']
        self.account_df['profit'] = self.account_df['balance'] - self.account_df['balance'].shift(fill_value=init_balance)  # 每日收益
        self.account_df['is_profit'] = np.where(self.account_df['profit'] > 0, 1, 0)  # 是否收益
        self.account_df['is_loss'] = np.where(self.account_df['profit'] < 0, 1, 0)  # 是否亏损
        self.account_df['daily_yield'] = self.account_df['balance'] / self.account_df['balance'].shift(fill_value=init_balance) - 1  # 每日收益率
        self.account_df['max_balance'] = self.account_df['balance'].cummax()  # 当前单日最大权益
        self.account_df['drawdown'] = (self.account_df['max_balance'] - self.account_df['balance']) / self.account_df['max_balance']  # 回撤
        _ror = self.account_df.iloc[-1]['balance'] / self.account_df.iloc[0]['pre_balance']
        return {
            "start_date": self.account_df.iloc[0]["date"],
            "end_date": self.account_df.iloc[-1]["date"],
            "init_balance": init_balance,
            "balance": balance,
            "start_balance": init_balance,
            "end_balance": balance,
            "ror": _ror - 1,  # 收益率
            "annual_yield": _ror ** (TRADING_DAYS_OF_YEAR / self.account_df.shape[0]) - 1,  # 年化收益率
            "trading_days": self.account_df.shape[0],  # 总交易天数
            "cum_profit_days": self.account_df['is_profit'].sum(),  # 累计盈利天数
            "cum_loss_days": self.account_df['is_loss'].sum(),  # 累计亏损天数
            "max_drawdown": self.account_df['drawdown'].max(),  # 最大回撤
            "commission": self.account_df['commission'].sum(),  # 总手续费
            "open_times": self.trade_df.loc[self.trade_df["offset1"] == "OPEN"].shape[0],  # 开仓次数
            "close_times": self.trade_df.loc[self.trade_df["offset1"] == "CLOSE"].shape[0],  # 平仓次数
            "daily_risk_ratio": self.account_df['risk_ratio'].mean(),  # 提供日均风险度
            "max_cont_profit_days": _cum_counts(self.account_df['is_profit']).max(),  # 最大连续盈利天数
            "max_cont_loss_days": _cum_counts(self.account_df['is_loss']).max(),  # 最大连续亏损天数
            "sharpe_ratio": get_sharp(self.account_df['daily_yield']),  # 年化夏普率
            "calmar_ratio": get_calmar(self.account_df['daily_yield'], self.account_df['drawdown'].max()),  # 年化卡玛比率
            "sortino_ratio": get_sortino(self.account_df['daily_yield']),  # 年化索提诺比率
            "tqsdk_punchline": self._get_tqsdk_punchlines(_ror - 1)
        }

    def _get_trades_stat_metrics(self):
        """
        根据成交手数计算 胜率，盈亏额比例
        self.quotes 主要需要合约乘数，用于计算盈亏额
        """
        trade_array = []
        for date in self.date_keys:
            for trade in self.trade_log[date]['trades']:
                # 每一行都是 1 手的成交记录
                trade_array.extend([{
                    "symbol": f"{trade['exchange_id']}.{trade['instrument_id']}",
                    "direction": trade["direction"],
                    "offset": "CLOSE" if trade["offset"] == "CLOSETODAY" else trade["offset"],
                    "price": trade["price"]
                } for i in range(trade['volume'])])
        trade_df = DataFrame(data=trade_array, columns=['symbol', 'direction', 'offset', 'price'])
        profit_volumes = 0  # 盈利手数
        loss_volumes = 0  # 亏损手数
        profit_value = 0  # 盈利额
        loss_value = 0  # 亏损额
        all_symbols = trade_df['symbol'].drop_duplicates()
        for symbol in all_symbols:
            for direction in ["BUY", "SELL"]:
                open_df = self._get_sub_df(trade_df, symbol, dir=direction, offset='OPEN')
                close_df = self._get_sub_df(trade_df, symbol, dir=("SELL" if direction == "BUY" else "BUY"), offset='CLOSE')
                close_df['profit'] = (close_df['price'] - open_df['price']) * (1 if direction == "BUY" else -1)
                profit_volumes += close_df.loc[close_df['profit'] >= 0].shape[0]  # 盈利手数
                loss_volumes += close_df.loc[close_df['profit'] < 0].shape[0]  # 亏损手数
                profit_value += close_df.loc[close_df['profit'] >= 0, 'profit'].sum() * self.quotes[symbol]['volume_multiple']
                loss_value += close_df.loc[close_df['profit'] < 0, 'profit'].sum() * self.quotes[symbol]['volume_multiple']
        winning_rate = profit_volumes / (profit_volumes + loss_volumes) if profit_volumes + loss_volumes else 0
        profit_pre_volume = profit_value / profit_volumes if profit_volumes else 0
        loss_pre_volume = loss_value / loss_volumes if loss_volumes else 0
        profit_loss_ratio = abs(profit_pre_volume / loss_pre_volume) if loss_pre_volume else float("inf")
        return {
            "profit_volumes": profit_volumes,
            "loss_volumes": loss_volumes,
            "profit_value": profit_value,
            "loss_value": loss_value,
            "winning_rate": winning_rate,
            "profit_loss_ratio": profit_loss_ratio
        }

    def _get_tqsdk_punchlines(self, ror):
        tqsdk_punchlines = [
            '幸好是模拟账户，不然你就亏完啦',
            '触底反弹,与其执迷修改参数，不如改变策略思路去天勤官网策略库进修',
            '越挫越勇，不如去天勤量化官网策略库进修',
            '不要灰心，少侠重新来过',
            '策略看来小有所成',
            '策略看来的得心应手',
            '策略看来春风得意，堪比当代索罗斯',
            '策略看来独孤求败，小心过拟合噢'
        ]
        ror_level = [i for i, k in enumerate([-1, -0.5, -0.2, 0, 0.2, 0.5, 1]) if ror < k]
        if len(ror_level) > 0:
            return tqsdk_punchlines[ror_level[0]]
        else:
            return tqsdk_punchlines[-1]

    def _get_sub_df(self, origin_df, symbol, dir, offset):
        df = origin_df.where(
            (origin_df['symbol'] == symbol) & (origin_df['offset'] == offset) & (origin_df['direction'] == dir))
        df.dropna(inplace=True)
        df.reset_index(drop=True, inplace=True)
        return df

    def metrics(self, **kwargs):
        self.default_metrics.update(kwargs)
        return [{
            self.report_id: {"metrics": self.default_metrics.copy()}
        }]

    def full(self):
        data = self.metrics()
        data += self.daily_balance()
        data += self.daily_profit()
        data += self.drawdown()
        data += self.sharp_rolling()
        data += self.sortino_rolling()
        # data += self.calmar_rolling()
        return data

    def daily_balance(self):
        """每日资金曲线"""
        return [{
            self.report_id: {
                "charts": {
                    "daily_balance": {
                        "title": {
                            "left": 'center',
                            "text": "每日账户资金"
                        },
                        "xAxis": {
                            "type": 'category',
                            "data": self.account_df['date'].to_dict()
                        },
                        "yAxis": {
                            "type": 'value',
                            "min": 'dataMin',
                            "max": 'dataMax',
                        },
                        "series": {
                            "0": {
                                "data": self.account_df['balance'].map(lambda x: '%.2f' % x).to_dict(),
                                "type": 'line'
                            }
                        }
                    }
                }
            }
        }]

    def daily_profit(self):
        """每日盈亏"""
        profit = Series(np.where(self.account_df['profit'] >= 0, self.account_df['profit'], float('nan')))  # 收益
        loss = Series(np.where(self.account_df['profit'] < 0, self.account_df['profit'], float('nan')))  # 亏损
        return [{
            self.report_id: {
                "charts": {
                    "daily_profit": {
                        "title": {
                            "left": 'center',
                            "text": "每日盈亏"
                        },
                        "xAxis": {
                            "type": 'category',
                            "data": self.account_df['date'].to_dict()
                        },
                        "yAxis": {
                            "type": 'value'
                        },
                        "series": {
                            "0": {
                                "data": profit.to_dict(),
                                "type": 'bar',
                                "itemStyle": {
                                    "color": "#ee6666"
                                },
                                "stack": 'one',
                            },
                            "1": {
                                "data": loss.to_dict(),
                                "type": 'bar',
                                "itemStyle": {
                                    "color": "#91cc75"
                                },
                                "stack": 'one',
                            }
                        }
                    }
                }
            }
        }]

    def drawdown(self):
        """回撤"""
        return [{
            self.report_id: {
                "charts": {
                    "drawdown": {
                        "title": {
                            "left": 'center',
                            "text": "回撤"
                        },
                        "xAxis": {
                            "type": 'category',
                            "data": self.account_df['date'].to_dict()
                        },
                        "yAxis": {
                            "type": 'value'
                        },
                        "series": {
                            "0": {
                                "data": self.account_df['drawdown'].to_dict(),
                                "type": 'line'
                            }
                        }
                    }
                }
            }
        }]

    def sharp_rolling(self):
        """滚动夏普比率图表"""
        rolling_sharp = self.account_df['daily_yield'].rolling(TRADING_DAYS_OF_MONTH).apply(get_sharp)
        return [{
            self.report_id: {
                "charts": {
                    "sharp_rolling": {
                        "title": {
                            "left": 'center',
                            "text": "滚动夏普比率图表"
                        },
                        "xAxis": {
                            "type": 'category',
                            "data": self.account_df['date'].to_dict()
                        },
                        "yAxis": {
                            "type": 'value'
                        },
                        "series": {
                            "0": {
                                "data": rolling_sharp.to_dict(),
                                "type": 'line'
                            }
                        }
                    }
                }
            }
        }]

    def sortino_rolling(self):
        """滚动索提诺比率图表"""
        rolling_sortino = self.account_df['daily_yield'].rolling(TRADING_DAYS_OF_MONTH).apply(get_sortino)
        return [{
            self.report_id: {
                "charts": {
                    "sortino_rolling": {
                        "title": {
                            "left": 'center',
                            "text": "滚动索提诺比率图表"
                        },
                        "xAxis": {
                            "type": 'category',
                            "data": self.account_df['date'].to_dict()
                        },
                        "yAxis": {
                            "type": 'value'
                        },
                        "series": {
                            "0": {
                                "name": "滚动索提诺比率图表",
                                "data": rolling_sortino.to_dict(),
                                "type": 'line'
                            }
                        }
                    }
                }
            }
        }]

    def calmar_rolling(self):
        """滚动卡玛比率图表"""
        rolling_calmar = self.account_df['daily_yield'].rolling(TRADING_DAYS_OF_MONTH).apply(
            lambda x: get_calmar(x, self.account_df.loc[x.index]['drawdown'].max()))
        return [{
            self.report_id: {
                "charts": {
                    "calmar_rolling": {
                        "title": {
                            "left": 'center',
                            "text": "滚动卡玛比率图表"
                        },
                        "xAxis": {
                            "type": 'category',
                            "data": self.account_df['date'].to_dict()
                        },
                        "yAxis": {
                            "type": 'value'
                        },
                        "series": {
                            "0": {
                                "data": rolling_calmar.to_dict(),
                                "type": 'line'
                            },
                        }
                    }
                }
            }
        }]



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = "Ringo"

'''
价格动量 策略 (难度：初级)
参考: https://www.shinnytech.com/blog/momentum-strategy/
注: 该示例策略仅用于功能示范, 实盘时请根据自己的策略/经验进行修改
'''

from tqsdk import TqApi, TqAuth, TargetPosTask

# 设置指定合约,获取N条K线计算价格动量
SYMBOL = "SHFE.au2012"
N = 15

api = TqApi(auth=TqAuth("信易账户", "账户密码"))
klines = api.get_kline_serial(SYMBOL, 60 * 60 * 24, N)
quote = api.get_quote(SYMBOL)
target_pos = TargetPosTask(api, SYMBOL)
position = api.get_position(SYMBOL)


def AR(kline1):
    """价格动量函数AR，以前N-1日K线计算价格动量ar"""
    spread_ho = sum(kline1.high[:-1] - kline1.open[:-1])
    spread_oc = sum(kline1.open[:-1] - kline1.low[:-1])
    # spread_oc 为0时，设置为最小价格跳动值
    if spread_oc == 0:
        spread_oc = quote.price_tick
    ar = (spread_ho / spread_oc) * 100
    return ar


ar = AR(klines)
print("策略开始启动")

while True:
    api.wait_update()
    # 生成新K线时，重新计算价格动量值ar
    if api.is_changing(klines.iloc[-1], "datetime"):
        ar = AR(klines)
        print("价格动量是：", ar)
    # 每次最新价发生变动时，重新进行判断
    if api.is_changing(quote, "last_price"):
        # 开仓策略
        if position.pos_long == 0 and position.pos_short == 0:
            # 如果ar大于110并且小于150，开多仓
            if 110 < ar < 150:
                print("价值动量超过110，小于150，做多")
                target_pos.set_target_volume(100)
            # 如果ar大于50，小于90，开空仓
            elif 50 < ar < 90:
                print("价值动量大于50，小于90，做空")
                target_pos.set_target_volume(-100)

        # 止损策略，多头下当前ar值小于90则平仓止损，空头下当前ar值大于110则平仓止损
        elif (position.pos_long > 0 and ar < 90) or (position.pos_short > 0 and ar > 110):
            print("止损平仓")
            target_pos.set_target_volume(0)



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'limin'

from tqsdk import TqApi, TqAuth

'''
画图示例: 在主图中画信号线及文字标注
注意: 画图示例中用到的数据不含有实际意义，请根据自己的实际策略情况进行修改
'''

api = TqApi(web_gui=True, auth=TqAuth("信易账户", "账户密码"))  # web_gui=True, 开启使用 web 界面查看绘图结果的功能
klines = api.get_kline_serial("SHFE.rb2105", 300)

# 示例1: 在主图中最后一根K线上画射线以标注需要的信号
api.draw_line(klines, -1, klines.iloc[-1].close, -1, klines.iloc[-1].high, line_type="SEG", color=0xFFFF9900, width=3)

# 示例2: 绘制字符串
api.draw_text(klines, "信号1", x=-1, y=klines.iloc[-1].high + 5, color=0xFFFF3333)

# 示例3: 给主图最后5根K线加一个方框
api.draw_box(klines, x1=-5, y1=klines.iloc[-5]["high"], x2=-1, y2=klines.iloc[-1]["low"], width=1, color=0xFF0000FF,
             bg_color=0x7000FF00)

# 由于需要在浏览器中查看绘图结果，因此程序不能退出
while True:
    api.wait_update()



# !usr/bin/env python3
# -*- coding:utf-8 -*-
__author__ = 'yanqiong'

import datetime
import os


DEBUG_DIR = os.path.join(os.path.expanduser('~'), ".tqsdk/logs")


def _get_log_name():
    """返回默认 debug 文件生成的位置"""
    if not os.path.exists(DEBUG_DIR):
        os.makedirs(os.path.join(os.path.expanduser('~'), ".tqsdk/logs"), exist_ok=True)
    return os.path.join(DEBUG_DIR, f"{datetime.datetime.now().strftime('%Y%m%d%H%M%S%f')}-{os.getpid()}.log")


def _clear_logs():
    """清除最后修改时间是 n 天前的日志"""
    if not os.path.exists(DEBUG_DIR):
        return
    n = os.getenv("TQ_SAVE_LOG_DAYS", 30)
    dt = datetime.datetime.now() - datetime.timedelta(days=int(n))
    for log in os.listdir(DEBUG_DIR):
        path = os.path.join(DEBUG_DIR, log)
        try:
            if datetime.datetime.fromtimestamp(os.stat(path).st_mtime) < dt:
                os.remove(path)
        except:
            pass  # 忽略抛错



#!usr/bin/env python3
# -*- coding:utf-8 -*-
__author__ = 'yanqiong'

import base64
import ctypes
import hashlib
import logging
import os
from pathlib import Path
import sys
import uuid
from typing import Optional

from tqsdk.tradeable.otg.base_otg import BaseOtg
from tqsdk.tradeable.mixin import FutureMixin


class TqAccount(BaseOtg, FutureMixin):
    """天勤实盘账户类"""

    def __init__(self, broker_id: str, account_id: str, password: str, front_broker: Optional[str] = None,
                 front_url: Optional[str] = None, td_url: Optional[str] = None, **kwargs) -> None:
        """
        创建天勤实盘账户实例

        Args:
            broker_id (str): 期货公司，支持的期货公司列表 https://www.shinnytech.com/blog/tq-support-broker/

            account_id (str): 帐号

            password (str): 密码

            td_url(str): [可选]用于指定账户连接的交易服务器地址, eg: "tcp://1.2.3.4:1234/"
        """
        if bool(front_broker) != bool(front_url):
            raise Exception("front_broker 和 front_url 参数需同时填写")
        self._front_broker = front_broker
        self._front_url = front_url
        self._app_id = "SHINNY_TQ_1.0"

        account_type = kwargs["account_type"] if "account_type" in kwargs else "FUTURE"
        if account_type == "SPOT":
            raise Exception("account_type 账户类型指定错误，目前只支持 FUTURE")
        kwargs.pop("account_type", None)
        if len(kwargs) > 0:
            raise TypeError(f"不支持以下参数 {[kwargs.keys()]}")

        super(TqAccount, self).__init__(broker_id, account_id, password, td_url)

    def _get_account_key(self):
        s = self._broker_id + self._account_id
        s += self._front_broker if self._front_broker else ""
        s += self._front_url if self._front_url else ""
        s += self._td_url if self._td_url else ""
        return hashlib.md5(s.encode('utf-8')).hexdigest()

    @property
    def _account_info(self):
        info = super(TqAccount, self)._account_info
        info.update({
            "account_type": self._account_type
        })
        return info

    def _get_system_info(self):
        try:
            l = ctypes.c_int(344)
            buf = ctypes.create_string_buffer(l.value)
            path = Path(__file__, '../../../ctpse')
            lib_path = path.resolve()  # Make the path absolute, resolving any symlinks. A new path object is returned
            if sys.platform.startswith("win") or sys.platform.startswith("linux"):
                if sys.platform.startswith("win"):
                    if ctypes.sizeof(ctypes.c_voidp) == 4:
                        selib = ctypes.cdll.LoadLibrary(os.path.join(lib_path, "WinDataCollect32.dll"))
                        ret = getattr(selib, "?CTP_GetSystemInfo@@YAHPADAAH@Z")(buf, ctypes.byref(l))
                    else:
                        selib = ctypes.cdll.LoadLibrary(os.path.join(lib_path, "WinDataCollect64.dll"))
                        ret = getattr(selib, "?CTP_GetSystemInfo@@YAHPEADAEAH@Z")(buf, ctypes.byref(l))
                else:
                    selib = ctypes.cdll.LoadLibrary(os.path.join(lib_path, "LinuxDataCollect64.so"))
                    ret = selib._Z17CTP_GetSystemInfoPcRi(buf, ctypes.byref(l))
                if ret == 0:
                    return base64.b64encode(buf.raw[:l.value]).decode("utf-8")
                else:
                    raise Exception("错误码: %d" % ret)
            else:
                logging.getLogger("TqApi.TqAccount").debug("ctpse error", error="不支持该平台")
        except Exception as e:
            self._api._print(f"采集穿透式监管客户端信息失败: {e}", level="ERROR")
            logging.getLogger("TqApi.TqAccount").error("ctpse error", error=e)
        return ""

    async def _send_login_pack(self):
        req = {
            "aid": "req_login",
            "bid": self._broker_id,
            "user_name": self._account_id,
            "password": self._password,
        }
        mac = f"{uuid.getnode():012X}"
        req["client_mac_address"] = "-".join([mac[e:e + 2] for e in range(0, 11, 2)])
        system_info = self._get_system_info()
        if system_info:
            req["client_app_id"] = self._app_id
            req["client_system_info"] = system_info
        if self._front_broker:
            req["broker_id"] = self._front_broker
            req["front"] = self._front_url
        await self._td_send_chan.send(req)
        await self._td_send_chan.send({
            "aid": "confirm_settlement"
        })  # 自动发送确认结算单



#!usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'mayanqiong'


from tqsdk.tradeable.otg.base_otg import BaseOtg
from tqsdk.tradeable.otg import TqAccount, TqKq, TqKqStock
from tqsdk.tradeable.sim.basesim import BaseSim
from tqsdk.tradeable.sim import TqSim, TqSimStock



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'


"""
时间帮助函数，根据回测/实盘获取不同的当前时间

模块内部创建使用
"""

import time


class TqDatetimeState:

    def __init__(self):
        self.data_ready = False  # 是否收到了第一个 mdhis_more_data
        self.tqsdk_backtest = {}

    def get_current_dt(self):
        """返回当前 nano timestamp"""
        if self.tqsdk_backtest:
            return self.tqsdk_backtest.get('current_dt')
        else:
            return int(time.time() * 1e9)

    def update_state(self, diff):
        self.tqsdk_backtest.update(diff.get('_tqsdk_backtest', {}))
        if not self.data_ready and diff.get('mdhis_more_data', True) is False:
            self.data_ready = True




#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'chengzhi'

from tqsdk import TqApi, TqAuth

api = TqApi(auth=TqAuth("信易账户", "账户密码"))
# 获得 m2105 的持仓引用，当持仓有变化时 position 中的字段会对应更新
position = api.get_position("DCE.m2105")
# 获得资金账户引用，当账户有变化时 account 中的字段会对应更新
account = api.get_account()
# 下单并返回委托单的引用，当该委托单有变化时 order 中的字段会对应更新
order = api.insert_order(symbol="DCE.m2105", direction="BUY", offset="OPEN", volume=5, limit_price=2750)

while True:
    api.wait_update()
    if api.is_changing(order, ["status", "volume_orign", "volume_left"]):
        print("单状态: %s, 已成交: %d 手" % (order.status, order.volume_orign - order.volume_left))
    if api.is_changing(position, "pos_long_today"):
        print("今多头: %d 手" % (position.pos_long_today))
    if api.is_changing(account, "available"):
        print("可用资金: %.2f" % (account.available))



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'chengzhi'

from tqsdk import TqApi, TqAuth, TargetPosTask

'''
价差回归
当近月-远月的价差大于250时做空近月，做多远月
当价差小于200时平仓
'''
api = TqApi(auth=TqAuth("信易账户", "账户密码"))
quote_near = api.get_quote("SHFE.rb2104")
quote_deferred = api.get_quote("SHFE.rb2105")
# 创建 rb2104 的目标持仓 task，该 task 负责调整 rb2104 的仓位到指定的目标仓位
target_pos_near = TargetPosTask(api, "SHFE.rb2104")
# 创建 rb2105 的目标持仓 task，该 task 负责调整 rb2105 的仓位到指定的目标仓位
target_pos_deferred = TargetPosTask(api, "SHFE.rb2105")

while True:
    api.wait_update()
    if api.is_changing(quote_near) or api.is_changing(quote_deferred):
        spread = quote_near.last_price - quote_deferred.last_price
        print("当前价差:", spread)
        if spread > 250:
            print("目标持仓: 空近月，多远月")
            # 设置目标持仓为正数表示多头，负数表示空头，0表示空仓
            target_pos_near.set_target_volume(-1)
            target_pos_deferred.set_target_volume(1)
        elif spread < 200:
            print("目标持仓: 空仓")
            target_pos_near.set_target_volume(0)
            target_pos_deferred.set_target_volume(0)



#!usr/bin/env python3
# -*- coding:utf-8 -*-
__author__ = 'yanqiong'

import asyncio
import json
import random
import ssl
import time
import warnings
from abc import abstractmethod
from datetime import datetime
from logging import Logger
from queue import Queue
from typing import Optional

import certifi
import websockets
from shinny_structlog import ShinnyLoggerAdapter

from tqsdk.diff import _merge_diff, _get_obj
from tqsdk.entity import Entity
from tqsdk.exceptions import TqBacktestPermissionError
from tqsdk.utils import _generate_uuid

"""
优化代码结构，修改为

TqConnect 
（负责连接 websocket 连接，从服务器收到数据发回下游，从下游收到指令包发给上游，生成连接建立、连接断开的通知发给下游）
 ｜  ｜
TqReconnectHandler 
(连通上下游，记录重连发生时需要重新发送的数据，在发生重连时，暂停接受下游数据、暂停转发上游数据到下游，直到从上游收到的数据集是完整数据截面，继续恢复工作)
 ｜  ｜
 xxxxxx
 ｜  ｜
 api
"""


class ReconnectTimer(object):

    def __init__(self):
        # 记录最大的下次重连的时间, 所有的 ws 连接，共用一个下次发起重连的时间，这个时间只会不断增大
        self.timer = time.time() + random.uniform(10, 20)

    def set_count(self, count):
        if self.timer < time.time():
            seconds = min(2 ** count, 64) * 10  # 最大是在 1280s ～ 2560s 之间
            self.timer = time.time() + random.uniform(seconds, seconds * 2)


class TqStreamReader(asyncio.StreamReader):

    def __init__(self, *args, **kwargs):
        super(TqStreamReader, self).__init__(*args, **kwargs)
        self._start_read_message = None
        self._read_size = 0

    async def readexactly(self, n):
        data = await super(TqStreamReader, self).readexactly(n)
        if not self._start_read_message:
            self._start_read_message = time.time()
        self._read_size += n
        return data


class TqWebSocketClientProtocol(websockets.WebSocketClientProtocol):

    def __init__(self, *args, **kwargs):
        super(TqWebSocketClientProtocol, self).__init__(*args, **kwargs)
        self.reader = TqStreamReader(limit=self.read_limit // 2, loop=self.loop)

    async def handshake(self, *args, **kwargs) -> None:
        try:
            await super(TqWebSocketClientProtocol, self).handshake(*args, **kwargs)
        except websockets.exceptions.InvalidStatusCode as e:
            for h_key, h_value in self.response_headers.items():
                if h_key == 'x-shinny-auth-check' and h_value == 'Backtest Permission Denied':
                    raise TqBacktestPermissionError(
                        "免费账户每日可以回测3次，今日暂无回测权限，需要购买专业版本后使用。升级网址：https://account.shinnytech.com") from None
            raise

    async def read_message(self):
        message = await super().read_message()
        self.reader._start_read_message = None
        self.reader._read_size = 0
        return message


class TqConnect(object):
    """用于与 websockets 服务器通讯"""

    def __init__(self, logger, conn_id: Optional[str] = None) -> None:
        """
        创建 TqConnect 实例
        """
        self._conn_id = conn_id if conn_id else _generate_uuid()
        self._logger = logger
        if isinstance(logger, Logger):
            self._logger = ShinnyLoggerAdapter(logger, conn_id=self._conn_id)
        elif isinstance(logger, ShinnyLoggerAdapter):
            self._logger = logger.bind(conn_id=self._conn_id)
        self._first_connect = True
        self._keywords = {"max_size": None}

    async def _run(self, api, url, send_chan, recv_chan):
        """启动websocket客户端"""
        self._api = api
        # 调整代码位置，方便 monkey patch
        self._ins_list_max_length = 100000  # subscribe_quote 最大长度
        self._subscribed_per_seconds = 100  # 每秒 subscribe_quote 请求次数限制
        self._subscribed_queue = Queue(self._subscribed_per_seconds)

        self._keywords["extra_headers"] = self._api._base_headers
        self._keywords["create_protocol"] = TqWebSocketClientProtocol
        if url.startswith("wss://"):
            ssl_context = ssl.create_default_context()
            ssl_context.load_verify_locations(certifi.where())
            self._keywords["ssl"] = ssl_context
        count = 0
        while True:
            try:
                if not self._first_connect:
                    notify_id = _generate_uuid()
                    notify = {
                        "type": "MESSAGE",
                        "level": "WARNING",
                        "code": 2019112910,
                        "conn_id": self._conn_id,
                        "content": f"开始与 {url} 的重新建立网络连接",
                        "url": url
                    }
                    self._logger.debug("websocket connection connecting")
                    await recv_chan.send({
                        "aid": "rtn_data",
                        "data": [{
                            "notify": {
                                notify_id: notify
                            }
                        }]
                    })
                async with websockets.connect(url, **self._keywords) as client:
                    # 发送网络连接建立的通知，code = 2019112901
                    notify_id = _generate_uuid()
                    notify = {
                        "type": "MESSAGE",
                        "level": "INFO",
                        "code": 2019112901,
                        "conn_id": self._conn_id,
                        "content": "与 %s 的网络连接已建立" % url,
                        "url": url
                    }
                    if not self._first_connect:  # 如果不是第一次连接, 即为重连
                        # 发送网络连接重新建立的通知，code = 2019112902
                        notify["code"] = 2019112902
                        notify["level"] = "WARNING"
                        notify["content"] = "与 %s 的网络连接已恢复" % url
                        self._logger.debug("websocket reconnected")
                    else:
                        self._logger.debug("websocket connected")
                    # 发送网络连接建立的通知，code = 2019112901 or 2019112902，这里区分了第一次连接和重连
                    await self._api._wait_until_idle()
                    await recv_chan.send({
                        "aid": "rtn_data",
                        "data": [{
                            "notify": {
                                notify_id: notify
                            }
                        }]
                    })
                    count = 0
                    self._api._reconnect_timer.set_count(count)
                    send_task = self._api.create_task(self._send_handler(send_chan, client))
                    try:
                        async for msg in client:
                            pack = json.loads(msg)
                            await self._api._wait_until_idle()
                            self._logger.debug("websocket received data", pack=msg)
                            await recv_chan.send(pack)
                    finally:
                        self._logger.debug("websocket connection info", current_time=time.time(),
                                           start_read_message=client.reader._start_read_message,
                                           read_size=client.reader._read_size)
                        send_task.cancel()
                        await send_task
            # 希望做到的效果是遇到网络问题可以断线重连, 但是可能抛出的例外太多了(TimeoutError,socket.gaierror等), 又没有文档或工具可以理出 try 代码中所有可能遇到的例外
            # 而这里的 except 又需要处理所有子函数及子函数的子函数等等可能抛出的例外, 因此这里只能遇到问题之后再补, 并且无法避免 false positive 和 false negative
            except (websockets.exceptions.ConnectionClosed, websockets.exceptions.InvalidStatusCode,
                    websockets.exceptions.InvalidState, websockets.exceptions.ProtocolError, OSError, EOFError,
                    TqBacktestPermissionError) as e:
                in_ops_time = datetime.now().hour == 19 and 0 <= datetime.now().minute <= 30
                # 发送网络连接断开的通知，code = 2019112911
                notify_id = _generate_uuid()
                notify = {
                    "type": "MESSAGE",
                    "level": "WARNING",
                    "code": 2019112911,
                    "conn_id": self._conn_id,
                    "content": f"与 {url} 的网络连接断开，请检查客户端及网络是否正常",
                    "url": url
                }
                if in_ops_time:
                    notify['content'] += '，每日 19:00-19:30 为日常运维时间，请稍后再试'
                self._logger.debug("websocket connection closed", error=str(e))
                await recv_chan.send({
                    "aid": "rtn_data",
                    "data": [{
                        "notify": {
                            notify_id: notify
                        }
                    }]
                })
                if isinstance(e, TqBacktestPermissionError):
                    # 如果错误类型是用户无回测权限，直接返回
                    raise
                if self._first_connect and in_ops_time:
                    raise Exception(f'与 {url} 的连接失败，每日 19:00-19:30 为日常运维时间，请稍后再试')
            finally:
                if self._first_connect:
                    self._first_connect = False
                # 下次重连的时间距离现在当前时间秒数，会等待相应的时间，否则立即发起重连
                sleep_seconds = self._api._reconnect_timer.timer - time.time()
                if sleep_seconds > 0:
                    await asyncio.sleep(sleep_seconds)
                count += 1
                self._api._reconnect_timer.set_count(count)

    async def _send_handler(self, send_chan, client):
        """websocket客户端数据发送协程"""
        try:
            async for pack in send_chan:
                if pack.get("aid") == "subscribe_quote":
                    if len(pack.get("ins_list", "")) > self._ins_list_max_length:
                        warnings.warn(f"订阅合约字符串总长度大于 {self._ins_list_max_length}，可能会引起服务器限制。", stacklevel=3)
                    if self._subscribed_queue.full():
                        first_time = self._subscribed_queue.get()
                        if time.time() - first_time < 1:
                            warnings.warn(f"1s 内订阅请求次数超过 {self._subscribed_per_seconds} 次，订阅多合约时推荐使用 api.get_quote_list 方法。", stacklevel=3)
                    self._subscribed_queue.put(time.time())
                msg = json.dumps(pack)
                await client.send(msg)
                self._logger.debug("websocket send data", pack=msg)
        except asyncio.CancelledError:  # 取消任务不抛出异常，不然等待者无法区分是该任务抛出的取消异常还是有人直接取消等待者
            pass


class TqReconnect(object):
    def __init__(self, logger):
        self._logger = logger
        self._resend_request = {}  # 重连时需要重发的请求
        self._un_processed = False  # 重连后尚未处理完标志
        self._pending_diffs = []
        self._data = Entity()
        self._data._instance_entity([])

    async def _run(self, api, api_send_chan, api_recv_chan, ws_send_chan, ws_recv_chan):
        self._api = api
        send_task = self._api.create_task(self._send_handler(api_send_chan, ws_send_chan))
        try:
            async for pack in ws_recv_chan:
                self._record_upper_data(pack)
                if self._un_processed:  # 处理重连后数据
                    pack_data = pack.get("data", [])
                    self._pending_diffs.extend(pack_data)
                    for d in pack_data:
                        # _merge_diff 之后， self._data 会用于判断是否接收到了完整截面数据
                        _merge_diff(self._data, d, self._api._prototype, persist=False, reduce_diff=False)
                    if self._is_all_received():
                        # 重连后收到完整数据截面
                        self._un_processed = False
                        pack = {
                            "aid": "rtn_data",
                            "data": self._pending_diffs
                        }
                        await api_recv_chan.send(pack)
                        self._logger = self._logger.bind(status=self._status)
                        self._logger.debug("data completed", pack=pack)
                    else:
                        await ws_send_chan.send({"aid": "peek_message"})
                        self._logger.debug("wait for data completed", pack={"aid": "peek_message"})
                else:
                    is_reconnected = False
                    for i in range(len(pack.get("data", []))):
                        for _, notify in pack["data"][i].get("notify", {}).items():
                            if notify["code"] == 2019112902:  # 重连建立
                                is_reconnected = True
                                self._un_processed = True
                                self._logger = self._logger.bind(status=self._status)
                                if i > 0:
                                    ws_send_chan.send_nowait({
                                        "aid": "rtn_data",
                                        "data": pack.get("data", [])[0:i]
                                    })
                                self._pending_diffs = pack.get("data", [])[i:]
                                break
                    if is_reconnected:
                        self._data = Entity()
                        self._data._instance_entity([])
                        for d in self._pending_diffs:
                            _merge_diff(self._data, d, self._api._prototype, persist=False, reduce_diff=False)
                        # 发送所有 resend_request
                        for msg in self._resend_request.values():
                            # 这里必须用 send_nowait 而不是 send，因为如果使用异步写法，在循环中，代码可能执行到 send_task, 可能会修改 _resend_request
                            ws_send_chan.send_nowait(msg)
                            self._logger.debug("resend request", pack=msg)
                        await ws_send_chan.send({"aid": "peek_message"})
                    else:
                        await api_recv_chan.send(pack)
        finally:
            send_task.cancel()
            await asyncio.gather(send_task, return_exceptions=True)

    async def _send_handler(self, api_send_chan, ws_send_chan):
        async for pack in api_send_chan:
            self._record_lower_data(pack)
            await ws_send_chan.send(pack)

    @property
    def _status(self):
        return "WAIT_FOR_COMPLETED" if self._un_processed else "READY"

    @abstractmethod
    def _is_all_received(self):
        """在重连后判断是否收到了全部的数据，可以继续处理后续的数据包"""
        pass

    def _record_upper_data(self, pack):
        """从上游收到的数据中，记录下重连时需要的数据"""
        pass

    def _record_lower_data(self, pack):
        """从下游收到的数据中，记录下重连时需要的数据"""
        pass


class MdReconnectHandler(TqReconnect):

    def _record_lower_data(self, pack):
        """从下游收到的数据中，记录下重连时需要的数据"""
        aid = pack.get("aid")
        if aid == "subscribe_quote":
            self._resend_request["subscribe_quote"] = pack
        elif aid == "set_chart":
            if pack["ins_list"]:
                self._resend_request[pack["chart_id"]] = pack
            else:
                self._resend_request.pop(pack["chart_id"], None)

    def _is_all_received(self):
        set_chart_packs = {k: v for k, v in self._resend_request.items() if v.get("aid") == "set_chart"}
        # 处理 seriesl(k线/tick)
        if not all([v.items() <= _get_obj(self._data, ["charts", k, "state"]).items()
                    for k, v in set_chart_packs.items()]):
            return False  # 如果当前请求还没收齐回应, 不应继续处理
        # 在接收并处理完成指令后, 此时发送给客户端的数据包中的 left_id或right_id 至少有一个不是-1 , 并且 mdhis_more_data是False；否则客户端需要继续等待数据完全发送
        if not all([(_get_obj(self._data, ["charts", k]).get("left_id", -1) != -1
                     or _get_obj(self._data, ["charts", k]).get("right_id", -1) != -1)
                    and not self._data.get("mdhis_more_data", True)
                    for k in set_chart_packs.keys()]):
            return False  # 如果当前所有数据未接收完全(定位信息还没收到, 或数据序列还没收到), 不应继续处理
        all_received = True  # 订阅K线数据完全接收标志
        for k, v in set_chart_packs.items():  # 判断已订阅的数据是否接收完全
            for symbol in v["ins_list"].split(","):
                if symbol:
                    path = ["klines", symbol, str(v["duration"])] if v["duration"] != 0 else ["ticks", symbol]
                    serial = _get_obj(self._data, path)
                    if serial.get("last_id", -1) == -1:
                        all_received = False
                        break
            if not all_received:
                break
        if not all_received:
            return False
        # 处理实时行情quote
        if self._data.get("ins_list", "") != self._resend_request.get("subscribe_quote", {}).get("ins_list", ""):
            return False  # 如果实时行情quote未接收完全, 不应继续处理
        return True


class TdReconnectHandler(TqReconnect):

    def __init__(self, logger):
        super().__init__(logger)
        self._pos_symbols = {}

    def _record_lower_data(self, pack):
        """从下游收到的数据中，记录下重连时需要的数据"""
        aid = pack.get("aid")
        if aid == "req_login":
            self._resend_request["req_login"] = pack
        elif aid == "confirm_settlement":
            self._resend_request["confirm_settlement"] = pack

    def _record_upper_data(self, pack):
        """从上游收到的数据中，记录下重连时需要的数据"""
        for d in pack.get("data", []):
            for user, trade_data in d.get("trade", {}).items():
                if user not in self._pos_symbols:
                    self._pos_symbols[user] = set()
                self._pos_symbols[user].update(trade_data.get("positions", {}).keys())

    def _is_all_received(self):
        """交易服务器只判断收到的 trade_more_data 是否为 False，作为收到完整数据截面的依据"""
        if not all([(not self._data.get("trade", {}).get(user, {}).get("trade_more_data", True))
                    for user in self._pos_symbols.keys()]):
            return False  # 如果交易数据未接收完全, 不应继续处理
        # 有可能重连之后，持仓比原有持仓减少，需要原有的数据集中删去减少的合约的持仓
        for user, trade_data in self._data.get("trade", {}).items():
            symbols = set(trade_data.get("positions", {}).keys())  # 当前真实持仓中的合约
            if self._pos_symbols.get(user, set()) > symbols:  # 如果此用户历史持仓中的合约比当前真实持仓中更多: 删除多余合约信息
                self._pending_diffs.append({
                    "trade": {
                        user: {
                            "positions": {symbol: None for symbol in (self._pos_symbols[user] - symbols)}
                        }
                    }
                })
        return True


class TsReconnectHandler(TqReconnect):

    def _record_lower_data(self, pack):
        """从下游收到的数据中，记录下重连时需要的数据"""
        aid = pack.get("aid")
        if aid == "subscribe_trading_status":
            self._resend_request["subscribe_trading_status"] = pack

    def _is_all_received(self):
        return True



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'

import asyncio

from tqsdk.objs import Quote


class TqSymbols(object):
    """
    天勤合约服务类
    """

    async def _run(self, api, sim_send_chan, sim_recv_chan, md_send_chan, md_recv_chan):
        """回测task"""
        self._api = api
        self._sim_send_chan = sim_send_chan
        self._sim_recv_chan = sim_recv_chan
        self._md_send_chan = md_send_chan
        self._md_recv_chan = md_recv_chan
        self._quotes_all_keys = set(Quote(None).keys())
        self._quotes_all_keys = self._quotes_all_keys.union({'margin', 'commission'})
        # 以下字段合约服务也会请求，但是不应该记在 quotes 中，quotes 中的这些字段应该有行情服务负责
        self._quotes_all_keys.difference_update({'pre_open_interest', 'pre_settlement', 'pre_close', 'upper_limit', 'lower_limit'})
        sim_task = self._api.create_task(self._sim_handler())
        try:
            async for pack in self._md_recv_chan:
                if pack.get("aid") == "rtn_data":
                    data = pack.setdefault("data", [])
                    # 对于收到的数据，全部转发给下游
                    # 对于合约服务信息，query_id 为 PYSDK_quote_xxx 开头的，一定是请求了合约的全部合约信息，需要转为 quotes 转发给下游
                    for d in data:
                        for query_id, query_result in d.get("symbols", {}).items():
                            if query_result:
                                if query_result.get("error", None):
                                    raise Exception(f"查询合约服务报错 {query_result['error']}")
                                elif query_id.startswith("PYSDK_quote"):
                                    data.append(
                                        {"quotes": self._api._symbols_to_quotes(query_result, self._quotes_all_keys)}
                                    )
                                    self._md_send_chan.send_nowait({
                                        "aid": "ins_query",
                                        "query_id": query_id,
                                        "query": ""
                                    })
                await self._sim_recv_chan.send(pack)
        finally:
            sim_task.cancel()
            await asyncio.gather(sim_task, return_exceptions=True)

    async def _sim_handler(self):
        # 下游发来的数据包，直接转发到上游
        async for pack in self._sim_send_chan:
            await self._md_send_chan.send(pack)



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
"""
交易日历处理
"""
__author__ = 'mayanqiong'

import os
from datetime import date, datetime
from typing import Union, List

import pandas as pd
import requests

rest_days_df = None
chinese_holidays_range = None


def _init_chinese_rest_days(headers=None):
    global rest_days_df, chinese_holidays_range
    if rest_days_df is None:
        url = os.getenv("TQ_CHINESE_HOLIDAY_URL", "https://files.shinnytech.com/shinny_chinese_holiday.json")
        rsp = requests.get(url, timeout=30, headers=headers)
        chinese_holidays = rsp.json()
        _first_day = date(int(chinese_holidays[0].split('-')[0]), 1, 1)  # 首个日期所在年份的第一天
        _last_day = date(int(chinese_holidays[-1].split('-')[0]), 12, 31)  # 截止日期所在年份的最后一天
        chinese_holidays_range = (_first_day, _last_day)
        rest_days_df = pd.DataFrame(data={'date': pd.Series(pd.to_datetime(chinese_holidays, format='%Y-%m-%d'))})
        rest_days_df['trading_restdays'] = False  # 节假日为 False
    return chinese_holidays_range


def _get_trading_calendar(start_dt: date, end_dt: date, headers=None):
    """
    获取一段时间内，每天是否是交易日

    :return: DataFrame
        date  trading
    2019-12-05  True
    2019-12-06  True
    2019-12-07  False
    2019-12-08  False
    2019-12-09  True
    """
    _init_chinese_rest_days(headers=headers)
    df = pd.DataFrame()
    df['date'] = pd.Series(pd.date_range(start=start_dt, end=end_dt, freq="D"))
    df['trading'] = df['date'].dt.dayofweek.lt(5)
    result = pd.merge(rest_days_df, df, sort=True, how="right", on="date")
    result.fillna(True, inplace=True)
    df['trading'] = result['trading'] & result['trading_restdays']
    return df


class TqContCalendar(object):

    """
    主连日历
    df:
        date   trading  KQ.m@DCE.a  KQ.m@DCE.eg
    2019-12-06    True  DCE.a2005  DCE.eg2001
    2019-12-09    True  DCE.a2005  DCE.eg2001
    2019-12-10    True  DCE.a2005  DCE.eg2005
    2019-12-11    True  DCE.a2005  DCE.eg2005
    2019-12-12    True  DCE.a2005  DCE.eg2005
    """

    continuous = None

    def __init__(self, start_dt: date, end_dt: date, symbols: Union[List[str], None] = None, headers=None) -> None:
        """
        初始化主连日历表
        :param date start_dt: 开始交易日日期
        :param date end_dt: 结束交易日日期
        :param list[str] symbols: 主连合约列表
        :return:
        """
        self.df = _get_trading_calendar(start_dt=start_dt, end_dt=end_dt, headers=headers)
        self.df = self.df.loc[self.df.trading, ['date']]  # 只保留交易日
        self.df.reset_index(inplace=True, drop=True)
        if TqContCalendar.continuous is None:
            rsp = requests.get(os.getenv("TQ_CONT_TABLE_URL", "https://files.shinnytech.com/continuous_table.json"), headers=headers)  # 下载历史主连合约信息
            rsp.raise_for_status()
            TqContCalendar.continuous = {f"KQ.m@{k}": v for k, v in rsp.json().items()}
        if symbols is not None:
            if not all([s in TqContCalendar.continuous.keys() for s in symbols]):
                raise Exception(f"参数错误，symbols={symbols} 中应该全部都是主连合约代码")
        symbols = TqContCalendar.continuous.keys() if symbols is None else symbols
        self.start_dt, self.end_dt = self.df.iloc[0].date, self.df.iloc[-1].date
        for s in symbols:
            self._ensure_cont_on_df(s)

    def _ensure_cont_on_df(self, cont):
        """将一个主连对应的标的填在 self.df 对应位置"""
        temp_df = pd.DataFrame(data=TqContCalendar.continuous[cont], columns=['date', 'underlying'])
        temp_df['date'] = pd.Series(pd.to_datetime(temp_df['date'], format='%Y%m%d'))
        merge_result = pd.merge(temp_df, self.df, sort=True, how="outer", on="date")
        merge_result.fillna(method="ffill", inplace=True)
        merge_result.fillna(value="", inplace=True)
        s = merge_result.loc[merge_result.date.ge(self.start_dt) & merge_result.date.le(self.end_dt), 'underlying']
        self.df[cont] = pd.Series(s.values)

    def _get_cont_underlying_on_date(self, dt: datetime):
        """返回某一交易日的全部主连"""
        df = self.df.loc[self.df.date.ge(dt), :]
        return df.iloc[0:1]



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'ringo'

from tqsdk import TqApi, TqAuth
from tqsdk.tafunc import time_to_datetime

'''
使用get_trading_status来判断合约是否进入交易状态来进行下单，该接口需要有天勤量化专业版资格才可使用
'''

api = TqApi(auth=TqAuth("信易账户", "账户密码"))
ts = api.get_trading_status("SHFE.cu2201")
print(ts.trade_status)

while True:
    api.wait_update()
    # 如果处于集合竞价状态则进行下单
    if ts.trade_status == "AUCTIONORDERING":
        order = api.insert_order("SHFE.cu2201", "BUY", "OPEN", 1, 71400)
        break
# insert_order指令会在下一次wait_update()发出
api.wait_update()

api.close()



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'

import asyncio
import functools
import sys
import time
from asyncio import Future
from typing import Optional, Coroutine


class TqBaseApi(object):
    """
    天勤基类，处理 EventLoop 相关的调用
    """

    def __init__(self, loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        """
        创建天勤接口实例

        Args:
            loop(asyncio.AbstractEventLoop): [可选] 使用指定的 IOLoop, 默认创建一个新的.
        """
        self._loop = asyncio.SelectorEventLoop() if loop is None else loop  # 创建一个新的 ioloop, 避免和其他框架/环境产生干扰
        self._event_rev, self._check_rev = 0, 0
        self._wait_idle_list = []  # 所有等待 loop idle 的 Future
        self._wait_timeout = False  # wait_update 是否触发超时
        self._tasks = set()  # 由api维护的所有根task，不包含子task，子task由其父task维护
        self._exceptions = []  # 由api维护的所有task抛出的例外
        # 回测需要行情和交易 lockstep, 而 asyncio 没有将内部的 _ready 队列暴露出来,
        # 因此 monkey patch call_soon 函数用来判断是否有任务等待执行
        self._loop.call_soon = functools.partial(self._call_soon, self._loop.call_soon)
        # Windows系统下asyncio不支持KeyboardInterrupt的临时补丁
        if sys.platform.startswith("win"):
            self._create_task(self._windows_patch())

    def _create_task(self, coro: Coroutine, _caller_api: bool = False) -> asyncio.Task:
        task = self._loop.create_task(coro)
        py_ver = sys.version_info
        current_task = asyncio.Task.current_task(loop=self._loop) if (py_ver.major == 3 and py_ver.minor < 7) else asyncio.current_task(loop=self._loop)
        if current_task is None or _caller_api:  # 由 api 创建的 task，需要 api 主动管理
            self._tasks.add(task)
            task.add_done_callback(self._on_task_done)
        return task

    def _call_soon(self, org_call_soon, callback, *args, **kargs):
        """ioloop.call_soon的补丁, 用来追踪是否有任务完成并等待执行"""
        self._event_rev += 1
        return org_call_soon(callback, *args, **kargs)

    def _run_once(self):
        """执行 ioloop 直到 ioloop.stop 被调用"""
        if not self._exceptions:
            self._loop.run_forever()
        if self._exceptions:
            raise self._exceptions.pop(0)

    def _run_until_idle(self, async_run=False):
        """执行 ioloop 直到没有待执行任务
        async_run is True 会从 _wait_idle_list 中取出等待的异步任务，保证同步代码优先于异步代码执行，
        只有被 _run_until_task_done 调用（即 api 等待 fetch_msg）时，async_run 会为 True
        """
        while self._check_rev != self._event_rev:
            check_handle = self._loop.call_soon(self._check_event, self._event_rev + 1)
            try:
                self._run_once()
            finally:
                check_handle.cancel()
        if len(self._wait_idle_list) > 0 and async_run:
            f = self._wait_idle_list.pop(0)  # 取出 list 中的第一个 Future
            f.set_result(None)  # f 返回

    async def _wait_until_idle(self):
        """等待 ioloop 执行到空闲时，才从网络连接处收数据包，在 TqConnect 类中使用"""
        f = Future()
        self._wait_idle_list.append(f)
        self._loop.stop()
        await f

    def _run_until_task_done(self, task: asyncio.Task, deadline=None):
        try:
            self._wait_timeout = False
            if deadline is not None:
                deadline_handle = self._loop.call_later(max(0, deadline - time.time()), self._set_wait_timeout)
            while not self._wait_timeout and not task.done():
                if len(self._wait_idle_list) == 0:
                    self._run_once()
                else:
                    self._run_until_idle(async_run=True)
        finally:
            if deadline is not None:
                deadline_handle.cancel()
            task.cancel()

    def _check_event(self, rev):
        self._check_rev = rev
        self._loop.stop()

    def _set_wait_timeout(self):
        self._wait_timeout = True
        self._loop.stop()

    def _on_task_done(self, task):
        """当由 api 维护的 task 执行完成后取出运行中遇到的例外并停止 ioloop"""
        try:
            exception = task.exception()
            if exception:
                self._exceptions.append(exception)
        except asyncio.CancelledError:
            pass
        finally:
            self._tasks.remove(task)
            self._loop.stop()

    async def _windows_patch(self):
        """Windows系统下asyncio不支持KeyboardInterrupt的临时补丁, 详见 https://bugs.python.org/issue23057"""
        while True:
            await asyncio.sleep(1)

    def _close(self) -> None:
        self._run_until_idle(async_run=False)  # 由于有的处于 ready 状态 task 可能需要报撤单, 因此一直运行到没有 ready 状态的 task
        for task in self._tasks:
            task.cancel()
        while self._tasks:  # 等待 task 执行完成
            self._run_once()
        self._loop.run_until_complete(self._loop.shutdown_asyncgens())
        self._loop.close()



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'limin'

'''
海龟策略 (难度：中级)
参考: https://www.shinnytech.com/blog/turtle/
注: 该示例策略仅用于功能示范, 实盘时请根据自己的策略/经验进行修改
'''

import json
import time
from tqsdk import TqApi, TqAuth, TargetPosTask
from tqsdk.ta import ATR


class Turtle:
    def __init__(self, symbol, account=None, auth=None, donchian_channel_open_position=20,
                 donchian_channel_stop_profit=10,
                 atr_day_length=20, max_risk_ratio=0.5):
        self.account = account  # 交易账号
        self.auth = auth  # 信易账户
        self.symbol = symbol  # 合约代码
        self.donchian_channel_open_position = donchian_channel_open_position  # 唐奇安通道的天数周期(开仓)
        self.donchian_channel_stop_profit = donchian_channel_stop_profit  # 唐奇安通道的天数周期(止盈)
        self.atr_day_length = atr_day_length  # ATR计算所用天数
        self.max_risk_ratio = max_risk_ratio  # 最高风险度
        self.state = {
            "position": 0,  # 本策略净持仓数(正数表示多头，负数表示空头，0表示空仓)
            "last_price": float("nan"),  # 上次调仓价
        }

        self.n = 0  # 平均真实波幅(N值)
        self.unit = 0  # 买卖单位
        self.donchian_channel_high = 0  # 唐奇安通道上轨
        self.donchian_channel_low = 0  # 唐奇安通道下轨

        self.api = TqApi(self.account, auth=self.auth)
        self.quote = self.api.get_quote(self.symbol)
        # 由于ATR是路径依赖函数，因此使用更长的数据序列进行计算以便使其值稳定下来
        kline_length = max(donchian_channel_open_position + 1, donchian_channel_stop_profit + 1, atr_day_length * 5)
        self.klines = self.api.get_kline_serial(self.symbol, 24 * 60 * 60, data_length=kline_length)
        self.account = self.api.get_account()
        self.target_pos = TargetPosTask(self.api, self.symbol)

    def recalc_paramter(self):
        # 平均真实波幅(N值)
        self.n = ATR(self.klines, self.atr_day_length)["atr"].iloc[-1]
        # 买卖单位
        self.unit = int((self.account.balance * 0.01) / (self.quote.volume_multiple * self.n))
        # 唐奇安通道上轨：前N个交易日的最高价
        self.donchian_channel_high = max(self.klines.high[-self.donchian_channel_open_position - 1:-1])
        # 唐奇安通道下轨：前N个交易日的最低价
        self.donchian_channel_low = min(self.klines.low[-self.donchian_channel_open_position - 1:-1])
        print("唐其安通道上下轨: %f, %f" % (self.donchian_channel_high, self.donchian_channel_low))
        return True

    def set_position(self, pos):
        self.state["position"] = pos
        self.state["last_price"] = self.quote["last_price"]
        self.target_pos.set_target_volume(self.state["position"])

    def try_open(self):
        """开仓策略"""
        while self.state["position"] == 0:
            self.api.wait_update()
            if self.api.is_changing(self.klines.iloc[-1], "datetime"):  # 如果产生新k线,则重新计算唐奇安通道及买卖单位
                self.recalc_paramter()
            if self.api.is_changing(self.quote, "last_price"):
                print("最新价: %f" % self.quote.last_price)
                if self.quote.last_price > self.donchian_channel_high:  # 当前价>唐奇安通道上轨，买入1个Unit；(持多仓)
                    print("当前价>唐奇安通道上轨，买入1个Unit(持多仓): %d 手" % self.unit)
                    self.set_position(self.state["position"] + self.unit)
                elif self.quote.last_price < self.donchian_channel_low:  # 当前价<唐奇安通道下轨，卖出1个Unit；(持空仓)
                    print("当前价<唐奇安通道下轨，卖出1个Unit(持空仓): %d 手" % self.unit)
                    self.set_position(self.state["position"] - self.unit)

    def try_close(self):
        """交易策略"""
        while self.state["position"] != 0:
            self.api.wait_update()
            if self.api.is_changing(self.quote, "last_price"):
                print("最新价: ", self.quote.last_price)
                if self.state["position"] > 0:  # 持多单
                    # 加仓策略: 如果是多仓且行情最新价在上一次建仓（或者加仓）的基础上又上涨了0.5N，就再加一个Unit的多仓,并且风险度在设定范围内(以防爆仓)
                    if self.quote.last_price >= self.state[
                        "last_price"] + 0.5 * self.n and self.account.risk_ratio <= self.max_risk_ratio:
                        print("加仓:加1个Unit的多仓")
                        self.set_position(self.state["position"] + self.unit)
                    # 止损策略: 如果是多仓且行情最新价在上一次建仓（或者加仓）的基础上又下跌了2N，就卖出全部头寸止损
                    elif self.quote.last_price <= self.state["last_price"] - 2 * self.n:
                        print("止损:卖出全部头寸")
                        self.set_position(0)
                    # 止盈策略: 如果是多仓且行情最新价跌破了10日唐奇安通道的下轨，就清空所有头寸结束策略,离场
                    if self.quote.last_price <= min(self.klines.low[-self.donchian_channel_stop_profit - 1:-1]):
                        print("止盈:清空所有头寸结束策略,离场")
                        self.set_position(0)

                elif self.state["position"] < 0:  # 持空单
                    # 加仓策略: 如果是空仓且行情最新价在上一次建仓（或者加仓）的基础上又下跌了0.5N，就再加一个Unit的空仓,并且风险度在设定范围内(以防爆仓)
                    if self.quote.last_price <= self.state[
                        "last_price"] - 0.5 * self.n and self.account.risk_ratio <= self.max_risk_ratio:
                        print("加仓:加1个Unit的空仓")
                        self.set_position(self.state["position"] - self.unit)
                    # 止损策略: 如果是空仓且行情最新价在上一次建仓（或者加仓）的基础上又上涨了2N，就平仓止损
                    elif self.quote.last_price >= self.state["last_price"] + 2 * self.n:
                        print("止损:卖出全部头寸")
                        self.set_position(0)
                    # 止盈策略: 如果是空仓且行情最新价升破了10日唐奇安通道的上轨，就清空所有头寸结束策略,离场
                    if self.quote.last_price >= max(self.klines.high[-self.donchian_channel_stop_profit - 1:-1]):
                        print("止盈:清空所有头寸结束策略,离场")
                        self.set_position(0)

    def strategy(self):
        """海龟策略"""
        print("等待K线及账户数据...")
        deadline = time.time() + 5
        while not self.recalc_paramter():
            if not self.api.wait_update(deadline=deadline):
                raise Exception("获取数据失败，请确认行情连接正常并已经登录交易账户")
        while True:
            self.try_open()
            self.try_close()


turtle = Turtle("SHFE.au2006")
print("策略开始运行")
try:
    turtle.state = json.load(open("turtle_state.json", "r"))  # 读取数据: 本策略目标净持仓数,上一次开仓价
except FileNotFoundError:
    pass
print("当前持仓数: %d, 上次调仓价: %f" % (turtle.state["position"], turtle.state["last_price"]))
try:
    turtle.strategy()
finally:
    turtle.api.close()
    json.dump(turtle.state, open("turtle_state.json", "w"))  # 保存数据



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'ringo'

from tqsdk import TqApi, TqAuth
from tqsdk.ta import OPTION_GREEKS

api = TqApi(auth=TqAuth("信易账户", "账户密码"))

# 获取指定期权行情
quote = api.get_quote("SHFE.cu2006C44000")

# 获取期权和其对应标的的多合约的 kline 数据
klines = api.get_kline_serial(["SHFE.cu2006C44000", "SHFE.cu2006"], 24 * 60 * 60, 30)

# 运行 OPTION_GREEKS 希腊值计算函数
greeks = OPTION_GREEKS(klines, quote, 0.025)

# 输出希腊字母
print(list(greeks["delta"]))
print(list(greeks["theta"]))
print(list(greeks["gamma"]))
print(list(greeks["vega"]))
print(list(greeks["rho"]))

api.close()



# -*- coding: utf-8 -*-
__author__ = 'yangyang'

import setuptools

# from py-spy/setup.py
try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel, get_platform

    class bdist_wheel(_bdist_wheel):

        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            # Mark us as not a pure python package (we have platform specific ctpse lib)
            if self.plat_name != "any":
                self.root_is_pure = False
                plat_name = (self.plat_name or get_platform()).replace('-', '_').replace('.', '_')
                if plat_name == "linux_x86_64" or plat_name == "manylinux1_x86_64":
                    self.distribution.package_data[""] = ["ctpse/LinuxDataCollect64.so"]
                elif plat_name == "win32":
                    self.distribution.package_data[""] = ["ctpse/WinDataCollect32.dll"]
                elif plat_name == "win_amd64":
                    self.distribution.package_data[""] = ["ctpse/WinDataCollect64.dll"]

        def get_tag(self):
            # this set's us up to build generic wheels.
            python, abi, plat = _bdist_wheel.get_tag(self)
            # We don't contain any python source
            python, abi = 'py3', 'none'
            return python, abi, plat
except ImportError:
    bdist_wheel = None

with open("README.md", mode="r", encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name='tqsdk',
    version="3.2.3",
    description='TianQin SDK',
    author='TianQin',
    author_email='tianqincn@gmail.com',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://www.shinnytech.com/tqsdk',
    packages=setuptools.find_packages(),
    zip_safe=False,
    python_requires='>=3.6.4',
    install_requires=["websockets>=8.1", "requests", "numpy", "pandas>=1.1.0", "scipy", "simplejson", "aiohttp",
                      "certifi", "pyjwt", "psutil", "shinny_structlog", "sgqlc"],
    cmdclass={'bdist_wheel': bdist_wheel},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True
)



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'limin'

'''
菲阿里四价 策略(日内突破策略, 在每日收盘前对所持合约进行平仓)
参考: https://www.shinnytech.com/blog/fairy-four-price/
注: 该示例策略仅用于功能示范, 实盘时请根据自己的策略/经验进行修改
'''

from tqsdk import TqApi, TqAuth, TargetPosTask
from datetime import datetime
import time

symbol = "SHFE.ni2012"  # 合约代码
close_hour, close_minute = 14, 50  # 平仓时间

api = TqApi(auth=TqAuth("信易账户", "账户密码"))  # 使用模拟帐号直连行情和交易服务器
quote = api.get_quote(symbol)  # 获取指定合约的盘口行情
klines = api.get_kline_serial(symbol, 24 * 60 * 60)  # 获取日线
position = api.get_position(symbol)  # 持仓信息
target_pos = TargetPosTask(api, symbol)  # 目标持仓

top_rail = klines.high.iloc[-2]  # 上轨: 昨日高点
bottom_rail = klines.low.iloc[-2]  # 下轨: 昨日低点
print("上轨:", top_rail, ",下轨:", bottom_rail, ",昨日收盘价:", klines.close.iloc[-2], ",今日开盘价:", klines.open.iloc[-1])

while True:
    api.wait_update()
    if api.is_changing(klines.iloc[-1], "datetime"):  # 如果产生一根新日线 (即到达下一个交易日): 重新获取上下轨
        top_rail = klines.high.iloc[-2]
        bottom_rail = klines.low.iloc[-2]
        print("上轨:", top_rail, ",下轨:", bottom_rail, ",昨日收盘价:", klines.close.iloc[-2], ",今日开盘价:", klines.open.iloc[-1])

    if api.is_changing(quote, "last_price"):  # 如果行情最新价发生变化
        print("当前最新价", quote.last_price)
        # 开仓突破
        if quote.last_price > top_rail and position.pos_long == 0:  # 如果价格突破上轨: 买入开仓
            print("最新价:", quote.last_price, ", 价格突破上轨,买入开仓")
            target_pos.set_target_volume(3)  # 设置目标持仓手数，将指定合约调整到目标头寸
        elif quote.last_price < bottom_rail and position.pos_short == 0:  # 如果价格跌破下轨: 卖出开仓
            print("最新价:", quote.last_price, ", 价格跌破下轨, 卖出开仓")
            target_pos.set_target_volume(-3)

        # 平仓止损: 当价格 向上突破上轨 或 向下突破下轨 后, 再次回破当日开盘价
        if (quote.highest > top_rail and quote.last_price <= quote.open) or (
                quote.lowest < bottom_rail and quote.last_price >= quote.open):
            print("平仓止损")
            target_pos.set_target_volume(0)

    if api.is_changing(quote, "datetime"):
        now_time = datetime.strptime(quote.datetime, "%Y-%m-%d %H:%M:%S.%f")  # 获取当前的行情时间
        if now_time.hour == close_hour and now_time.minute >= close_minute:  # 到达平仓时间: 平仓
            print("临近本交易日收盘: 平仓")
            target_pos.set_target_volume(0)
            deadline = time.time() + 60  # 设置截止时间为当前时间的60秒以后
            while api.wait_update(deadline=deadline):  # 等待60秒
                pass
            api.close()  # 关闭api
            break  # 退出while循环



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'chengzhi'

from tqsdk import TqApi, TqAuth

# 可以指定debug选项将调试信息写入指定的文件中
api = TqApi(debug="debug.log", auth=TqAuth("信易账户", "账户密码"))
quote = api.get_quote("CZCE.FG209")
print(quote.datetime, quote.last_price, quote.ask_price1, quote.ask_price2)

while True:
    # 调用 wait_update 等待业务信息发生变化，例如: 行情发生变化, 委托单状态变化, 发生成交等等
    # 注意：其他合约的行情的更新也会触发业务信息变化，因此下面使用 is_changing 判断 FG209 的行情是否有变化
    api.wait_update()
    # 如果 FG209 的任何字段有变化，is_changing就会返回 True
    if api.is_changing(quote):
        print("行情变化", quote)
    # 只有当 FG209 的最新价有变化，is_changing才会返回 True
    if api.is_changing(quote, "last_price"):
        print("最新价变化", quote.last_price)
    # 当 FG209 的买1价/买1量/卖1价/卖1量中任何一个有变化，is_changing都会返回 True
    if api.is_changing(quote, ["ask_price1", "ask_volume1", "bid_price1", "bid_volume1"]):
        print("盘口变化", quote.ask_price1, quote.ask_volume1, quote.bid_price1, quote.bid_volume1)



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'limin'

from tqsdk import TqApi, TqAuth
from tqsdk.ta import MA

'''
画图示例: 在主图中画指标线
注意: 画图示例中用到的数据不含有实际意义，请根据自己的实际策略情况进行修改
'''

api = TqApi(web_gui=True, auth=TqAuth("信易账户", "账户密码"))  # web_gui=True, 开启使用 web 界面查看绘图结果的功能
klines = api.get_kline_serial("SHFE.rb2105", 5)

# 画一次指标线
ma = MA(klines, 30)  # 使用 tqsdk 自带指标函数计算均线
klines["ma_MAIN"] = ma.ma  # 在主图中画一根默认颜色（红色）的 ma 指标线

# 由于需要在浏览器中查看绘图结果，因此程序不能退出
while True:
    api.wait_update()
